/**
 * 
 */
package sim.util;

import java.io.File;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;

/**
 * La classe <b>SFileSearch</b> permet la recherche d'un nom de <u>fichier</u>  partir d'un emplacement dans un ensemble de rpertoire 
 * et de sous-rpertoire de faon rcursif.  
 * 
 * @author Simon Vzina
 * @since 2015-04-02
 * @version 2018-02-23
 */
public class SFileSearch {

  //-------------
  // CONSANTES //
  //-------------
  
  /**
   * La constante <b>DEFAULT_MAX_DEPTH<b> correspond au niveau de profondeur maximal de recherche par rcursivit tant gal  {@value}.
   */
  private final static int DEFAULT_MAX_DEPTH = 8;
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>max_depth<b> correspond au niveau de profondeur de recherche par rcursivit.
   */
  private final int max_depth;
  
  /**
   * La variable 'file_name_to_search' correspond au fichier  recherche. Ce nom peut comprendre "rpertoire de localisation"/"nom du fichier".
   * S'il y a un nom de rpertoire d'inclu au nom du fichier, cette information sera ajoute  la variablle 'starting_subdirectory' lors de la recherche.
   */
  private final String file_name_to_search;     
  
  /**
   * La variable 'starting_subdirectory' correspond au nom de rpertoire  partir du quel nous allons dbuter la recherche du fichier.
   * Il est important de prciser que la recherche commencera <b>toujours</b>  partir du rpertoire <b>"user.dir"</b>.
   */
  private final String starting_subdirectory;  
  
  /**
   * La variable 'file_found_list' correspond  la liste des noms de fichier (rpertoire/nom_fichier) o le fichier en recherche a t trouv.
   */
  private final List<String> file_found_list;    
 
  /**
   * Constructeur d'un chercheur de fichier  partir du rpertoire "user.dir" o l'application est lance.
   *
   * @param file_name Le nom du fichier  chercher.
   * @throws SConstructorException Si le sous-rpertoire ne permet pas d'identifier un rpertoire valide.
   * @see System.getProperty("user.dir")
   */
  public SFileSearch(String file_name) throws SConstructorException
  {
    this("",file_name);
  }
  
  /**
   * Constructeur d'un chercheur de fichier  partir du rpertoire "user.dir" o l'application est lance
   * et d'un sous-rpertoire. Le recherche par rcursivit aura une profondeur maximal par dfaut.
   * 
   * @param subdirectory Le nom du sous-rpertoire o dbutera la recherche  partir du rpertoire "user.dir". 
   * @param file_name Le nom du fichier  chercher.
   * @throws SConstructorException Si le sous-rpertoire ne permet pas d'identifier un rpertoire valide.
   * @see System.getProperty("user.dir")
   */
  public SFileSearch(String subdirectory, String file_name) throws SConstructorException
  {
    this(subdirectory, file_name, DEFAULT_MAX_DEPTH);
  }
  
  /**
   * Constructeur d'un chercheur de fichier  partir du rpertoire "user.dir" o l'application est lance
   * et d'un sous-rpertoire.
   * 
   * @param subdirectory Le nom du sous-rpertoire o dbutera la recherche  partir du rpertoire "user.dir". 
   * @param file_name Le nom du fichier  chercher.
   * @throws SConstructorException Si le sous-rpertoire ne permet pas d'identifier un rpertoire valide.
   * @see System.getProperty("user.dir")
   */
  public SFileSearch(String subdirectory, String file_name, int depth) throws SConstructorException
  {
    // Vrification du niveau de profondeur.
    if(depth < 1)
      throw new SConstructorException("Erreur SFileSearch 001 : Le niveau de profondeur '" + depth + "' doit tre suprieur  0.");
    
    max_depth = depth;
    
    // Mettre le sous-rpertoire en minuscule
    //subdirectory = subdirectory.toLowerCase();
    
    // Obtenir le nom du fichier sans les informations de rpertoire de localisation du fichier.
    file_name_to_search = SStringUtil.getFileNameWithoutDirectory(file_name.toLowerCase());  
        
    // tablir le nom du rpertoire initiale de recherche dbutant par le rpertoire "user.dir" (l o l'application est lance)
    if(subdirectory.equals(""))
      this.starting_subdirectory = System.getProperty("user.dir");
    else
      this.starting_subdirectory = System.getProperty("user.dir") + SStringUtil.FILE_SEPARATOR_CARACTER + subdirectory;
    
    // Liste des fichiers trouvs, car il peut y avoir plusieurs fichiers avec le mme nom
    file_found_list = new ArrayList<String>();        
    
    // Cration du rpertoire o dbut la recherche
    File directory = new File(starting_subdirectory); 
    
    // Vrification de la validit du nom du rpertoire
    if(directory.isDirectory())
      search(directory, 0);        //dbut de la recherche
    else
      throw new SConstructorException("Erreur SFileSearch 001 : Le sous-rpertoire '" + subdirectory + "' n'est pas un rpertoire.");
  }
  
  /**
   * Mthode pour obtenir le nom du fichier  rechercher.
   * @return Le nom du fichier recherch.
   */
  public String getFileNameToSearch()
  {
    return file_name_to_search;
  }
 
  /**
   * Mthode pour dterminer si le fichier a t trouv. Cependant, il est possible que le nom du fichier ait t trouv plus d'une fois.
   * @return <b>true</b> si le fichier a t trouv et <b>false</b> sinon.
   */
  public boolean isFileFound()
  {
    return !file_found_list.isEmpty();
  }
  
  /**
   * Mthode pour dterminer si le fichier a t trouv <b>plus d'une fois</b>.
   * @return <b>true</b> si le fichier a t trouv plus d'une fois et <b>false</b> sinon. 
   */
  public boolean isManyFileFound()
  {
    return file_found_list.size() > 1;
  }
  
  /**
   * Mthode pour obtenir la liste des adresses o l'on a trouv le nom du fichier. La liste sera vide si le fichier n'a pas t trouv.
   * @return La liste des adresses o le fichier a t trouv.
   */
  public List<String> getFileFoundList()
  {
    return file_found_list;
  }
 
  /**
   * Mthode pour faire la recherche du fichier dsir de faon rcursive.  chaque fois que le fichier sera trouv,
   * l'adresse du fichier sera sauvegard dans la liste.
   * 
   * @param file Le rpertoire o la recherche est rendue.
   * @param depth Le niveau de profondeur de la rcursivit.
   */
  private void search(File file, int depth)
  {
    // Vrification du niveau de profondeur maximal du fichier en recherche
    if(depth < max_depth)
    {
      if(file.isDirectory())                  // Vrification si l'analyse se fait sur un rpertoire
        if(file.canRead())                    // Vrification si l'accs au rpertoire est possible
          for(File temp : file.listFiles())   // Itrer sur l'ensemble du contenu du rpertoire
            if(temp.isDirectory())            // Vrifier si le fichier itr est un rpertoire
              search(temp, depth + 1);        // Lecture rcursive du rpertoire de niveau suprieur
            else 
              if(getFileNameToSearch().equals(temp.getName().toLowerCase()))  // Si le fichier est trouv ! 
                file_found_list.add(temp.getAbsoluteFile().toString());    
    }
  }
  
}//fin classe SFileSearch
